@@ -3,7 +3,7 @@
 
 # ******************************************************************************
 #
-# Copyright (C) 2006-2007 Olivier Tilloy <olivier@tilloy.net>
+# Copyright (C) 2006-2008 Olivier Tilloy <olivier@tilloy.net>
 #
 # This file is part of the pyexiv2 distribution.
 #
@@ -270,6 +270,117 @@ def StringToTime(string):
 
 	return localTime
 
+class Rational:
+
+	"""
+	A class representing a rational number.
+	"""
+
+	def __init__(self, numerator, denominator):
+		"""
+		Constructor.
+
+		Construct a rational number from its numerator and its denominator.
+
+		Keyword arguments:
+		numerator -- the numerator
+		denominator -- the denominator (if zero, will raise a ZeroDivisionError)
+		"""
+		if int(denominator) == 0:
+			raise ZeroDivisionError('Denominator of a rational number cannot be zero')
+		self.numerator = long(numerator)
+		self.denominator = long(denominator)
+
+	def __eq__(self, other):
+		"""
+		Compare two rational numbers for equality.
+
+		Two rational numbers are equal if and only if their numerators are equal
+		and their denominators are equal.
+
+		Keyword arguments:
+		other -- the rational number to compare to self for equality
+		"""
+		return ((self.numerator == other.numerator) and
+		        (self.denominator == other.denominator))
+
+	def __str__(self):
+		"""
+		Return a string representation of the rational number.
+		"""
+		return str(self.numerator) + '/' + str(self.denominator)
+
+def StringToRational(string):
+	"""
+	Try to convert a string containing a rational number to a Rational object.
+
+	Try to convert a string containing a rational number to the corresponding
+	Rational object.
+	The conversion is done by matching a regular expression.
+	If the pattern does not match, the Rational object with numerator=0 and
+	denominator=1 is returned.
+
+	Keyword arguments:
+	string -- the string potentially containing a rational number
+	"""
+	pattern = re.compile("(-?[0-9]+)/(-?[1-9][0-9]*)")
+	match = pattern.match(string)
+	if match == None:
+		return Rational(0, 1)
+	else:
+		return Rational(*map(long, match.groups()))
+
+def ConvertToPythonType(tagFamily, tagType, tagValue):
+	"""
+	Types a tag value using Python's built-in types or modules.
+
+	Whenever possible, the value is typed using Python's built-in types or
+	modules such as date when the value represents a date (e.g. the IPTC tag
+	'Iptc.Application2.DateCreated').
+	For EXIF rational number, custom type pyexiv2.Rational is used.
+
+	Keyword arguments:
+	tagFamily -- the family of the tag ('Exif' or 'Iptc')
+	tagType -- the type of the tag as defined in the EXIF and IPTC specifications
+	tagValue -- the value of the tag as a raw string
+	"""
+	value = tagValue
+	if tagFamily == 'Exif':
+		if tagType == 'Byte':
+			pass
+		elif tagType == 'Ascii':
+			# try to guess if the value is a datetime
+			value = StringToDateTime(tagValue)
+		elif tagType == 'Short':
+			value = int(tagValue)
+		elif tagType == 'Long' or tagType == 'SLong':
+			value = long(tagValue)
+		elif tagType == 'Rational' or tagType == 'SRational':
+			value = StringToRational(tagValue)
+		elif tagType == 'Undefined':
+			# tagValue is a sequence of bytes whose codes are written as a
+			# string, each code being followed by a blank space (e.g.
+			# "48 50 50 49 " for "0221").
+			try:
+				value = UndefinedToString(tagValue)
+			except ValueError:
+				# Some tags such as "Exif.Photo.UserComment" are marked as
+				# Undefined but do not store their value as expected.
+				# This should fix bug #173387.
+				pass
+	elif tagFamily == 'Iptc':
+		if tagType == 'Short':
+			value = int(tagValue)
+		elif tagType == 'String':
+			pass
+		elif tagType == 'Date':
+			value = StringToDate(tagValue)
+		elif tagType == 'Time':
+			value = StringToTime(tagValue)
+		elif tagType == 'Undefined':
+			pass
+	return value
+
 class Image(libpyexiv2.Image):
 
 	"""
@@ -281,9 +392,13 @@ class Image(libpyexiv2.Image):
 	"""
 
 	def __init__(self, filename):
-		libpyexiv2.Image.__init__(self, str(filename))
+		if filename.__class__ is unicode:
+			filename = filename.encode('utf-8')
+		libpyexiv2.Image.__init__(self, filename)
 		self.__exifTagsDict = {}
 		self.__iptcTagsDict = {}
+		self.__exifCached = False
+		self.__iptcCached = False
 
 	def __getExifTagValue(self, key):
 		"""
@@ -298,44 +413,14 @@ class Image(libpyexiv2.Image):
 		key -- the EXIF key of the requested metadata tag
 		"""
 		tagType, tagValue = self.__getExifTag(key)
-		if tagType == 'Byte':
-			return tagValue
-		elif tagType == 'Ascii':
-			# try to guess if the value is a datetime
-			return StringToDateTime(tagValue)
-		elif tagType == 'Short':
-			# In case the tag value is malformed, containing several short
-			# values separated by spaces (e.g. "2 3 4 5", see bug #146323),
-			# keep only the first of these values.
-			return int(tagValue.split()[0])
-		elif tagType == 'Long' or tagType == 'SLong':
-			return long(tagValue.split()[0])
-		# for Rational and SRational types, we use tuples
-		# TODO: define a rational type?
-		elif tagType == 'Rational':
-			pattern = re.compile("([0-9]+)/([1-9][0-9]*)")
-			match = pattern.match(tagValue)
-			if match == None:
-				return long(0), long(1)
+		if tagType not in ('Byte', 'Ascii', 'Undefined'):
+			values = [ConvertToPythonType('Exif', tagType, x) for x in tagValue.split()]
+			if len(values) == 1:
+				return values[0]
 			else:
-				v = map(long, match.groups())
-				return v[0], v[1]
-		elif tagType == 'SRational':
-			pattern = re.compile("(-?[0-9]+)/(-?[1-9][0-9]*)")
-			match = pattern.match(tagValue)
-			if match == None:
-				return long(0), long(1)
-			else:
-				v = map(long, match.groups())
-				return v[0], v[1]
-		elif tagType == 'Undefined':
-			# tagValue is a sequence of bytes whose codes are written as a
-			# string, each code being followed by a blank space (e.g.
-			# "48 50 50 49 " for "0221").
-			# Note: in the case of tag "Exif.Photo.UserComment", it is better to
-			# call method __getExifTagToString() to obtain directly the value as
-			# a human-readable string.
-			return UndefinedToString(tagValue)
+				return tuple(values)
+		else:
+			return ConvertToPythonType('Exif', tagType, tagValue)
 
 	def __setExifTagValue(self, key, value):
 		"""
@@ -356,8 +441,8 @@ class Image(libpyexiv2.Image):
 			strVal = str(value)
 		elif valueType == datetime.datetime:
 			strVal = value.strftime('%Y:%m:%d %H:%M:%S')
-		elif valueType == tuple:
-			strVal = '%s/%s' % (str(value[0]), str(value[1]))
+		elif valueType == list or valueType == tuple:
+			strVal = ' '.join([str(x) for x in value])
 		else:
 			# Value must already be a string.
 			# Warning: no distinction is possible between values that really are
@@ -366,29 +451,8 @@ class Image(libpyexiv2.Image):
 			# a string correctly formatted, using utility function
 			# StringToUndefined().
 			strVal = str(value)
-		self.__setExifTag(key, strVal)
-
-	def __convertIptcTagValueToPythonType(self, (tagType, tagValue)):
-		"""
-		Types a tag value using Python's built-in types or modules.
-
-		Whenever possible, the value is typed using Python's built-in types or
-		modules such as date when the value represents a date (e.g. the IPTC tag
-		'Iptc.Application2.DateCreated').
-
-		Keyword arguments:
-		(tagType, tagValue) -- the type and the value of the IPTC tag as strings
-		"""
-		if tagType == 'Short':
-			return int(tagValue)
-		elif tagType == 'String':
-			return tagValue
-		elif tagType == 'Date':
-			return StringToDate(tagValue)
-		elif tagType == 'Time':
-			return StringToTime(tagValue)
-		elif tagType == 'Undefined':
-			return tagValue
+		typeName, oldValue = self.__setExifTag(key, strVal)
+		return typeName
 
 	def __getIptcTagValue(self, key):
 		"""
@@ -405,8 +469,7 @@ class Image(libpyexiv2.Image):
 		Keyword arguments:
 		key -- the IPTC key of the requested metadata tag
 		"""
-		typeValuesList = self.__getIptcTag(key)
-		return map(self.__convertIptcTagValueToPythonType, typeValuesList)
+		return [ConvertToPythonType('Iptc', *x) for x in self.__getIptcTag(key)]
 
 	def __setIptcTagValue(self, key, value, index=0):
 		"""
@@ -447,7 +510,8 @@ class Image(libpyexiv2.Image):
 			# strings (type 'String') and those that are of type 'Undefined'.
 			# FIXME: for tags of type 'Undefined', this does not seem to work...
 			strVal = str(value)
-		self.__setIptcTag(key, strVal, index)
+		typeName, oldValue = self.__setIptcTag(key, strVal, index)
+		return typeName
 
 	def __getitem__(self, key):
 		"""
@@ -527,11 +591,12 @@ class Image(libpyexiv2.Image):
 				if value.__class__ is datetime.datetime:
 					value = value.replace(microsecond=0)
 
-				self.__setExifTagValue(key, value)
-				self.__exifTagsDict[key] = value
+				typeName = self.__setExifTagValue(key, value)
+				self.__exifTagsDict[key] = ConvertToPythonType(tagFamily, typeName, str(value))
 			else:
 				self.__deleteExifTag(key)
-				del self.__exifTagsDict[key]
+				if self.__exifTagsDict.has_key(key):
+					del self.__exifTagsDict[key]
 		elif tagFamily == 'Iptc':
 			# The case of IPTC tags is a bit trickier since some tags are
 			# repeatable. To simplify the process, parameter 'value' is
@@ -555,8 +620,12 @@ class Image(libpyexiv2.Image):
 				if oldValues.__class__ is not tuple:
 					oldValues = (oldValues,)
 			except KeyError:
-				# The tag is not set yet
-				oldValues = ()
+				# The tag is not cached yet
+				try:
+					oldValues = self.__getitem__(key)
+				except KeyError:
+					# The tag is not set
+					oldValues = ()
 
 			# For time objects, microseconds are not supported by the IPTC
 			# specification, so truncate them if present.
@@ -568,7 +637,7 @@ class Image(libpyexiv2.Image):
 					tempNewValues.append(newValue)
 			newValues = tuple(tempNewValues)
 
-			# This loop processes the values one by one. There are n cases:
+			# This loop processes the values one by one. There are 3 cases:
 			#   * if the two tuples are of the exact same size, each item in
 			#     oldValues is replaced by its new value in newValues;
 			#   * if newValues is longer than oldValues, each item in oldValues
@@ -579,15 +648,19 @@ class Image(libpyexiv2.Image):
 			#     extra items in oldValues are deleted.
 			for i in xrange(max(len(oldValues), len(newValues))):
 				try:
-					self.__setIptcTagValue(key, newValues[i], i)
+					typeName = self.__setIptcTagValue(key, newValues[i], i)
 				except IndexError:
-					self.__deleteIptcTag(key, min(len(oldValues), len(newValues)))
+					try:
+						self.__deleteIptcTag(key, min(len(oldValues), len(newValues)))
+					except KeyError:
+						pass
 			if len(newValues) > 0:
 				if len(newValues) == 1:
 					newValues = newValues[0]
-				self.__iptcTagsDict[key] = newValues
+				self.__iptcTagsDict[key] = tuple([ConvertToPythonType(tagFamily, typeName, str(v)) for v in newValues])
 			else:
-				del self.__iptcTagsDict[key]
+				if self.__iptcTagsDict.has_key(key):
+					del self.__iptcTagsDict[key]
 		else:
 			raise IndexError("Invalid key `" + key + "'")
 
@@ -608,7 +681,8 @@ class Image(libpyexiv2.Image):
 		tagFamily = key[:4]
 		if tagFamily == 'Exif':
 			self.__deleteExifTag(key)
-			del self.__exifTagsDict[key]
+			if self.__exifTagsDict.has_key(key):
+				del self.__exifTagsDict[key]
 		elif tagFamily == 'Iptc':
 			try:
 				oldValues = self.__iptcTagsDict[key]
@@ -616,7 +690,8 @@ class Image(libpyexiv2.Image):
 				oldValues = self.__getIptcTag(key)
 			for i in xrange(len(oldValues)):
 				self.__deleteIptcTag(key, 0)
-			del self.__iptcTagsDict[key]
+			if self.__iptcTagsDict.has_key(key):
+				del self.__iptcTagsDict[key]
 		else:
 			raise IndexError("Invalid key `" + key + "'")
 
@@ -627,8 +702,10 @@ class Image(libpyexiv2.Image):
 		Read the values of all the EXIF tags in the image and cache them in an
 		internal dictionary so as to speed up subsequent accesses.
 		"""
-		for key in self.exifKeys():
-			self[key]
+		if not self.__exifCached:
+			for key in self.exifKeys():
+				self[key]
+			self.__exifCached = True
 
 	def cacheAllIptcTags(self):
 		"""
@@ -637,8 +714,10 @@ class Image(libpyexiv2.Image):
 		Read the values of all the IPTC tags in the image and cache them in an
 		internal dictionary so as to speed up subsequent accesses.
 		"""
-		for key in self.iptcKeys():
-			self[key]
+		if not self.__iptcCached:
+			for key in self.iptcKeys():
+				self[key]
+			self.__iptcCached = True
 
 	def interpretedExifValue(self, key):
 		"""
@@ -657,6 +736,22 @@ class Image(libpyexiv2.Image):
 		# This method was added as a requirement tracked by bug #147534
 		return self.__getExifTagToString(key)
 
+	def copyMetadataTo(self, destImage):
+		"""
+		Duplicate all the tags and the comment from this image to another one.
+
+		Read all the values of the EXIF and IPTC tags and the comment and write
+		them back to the new image.
+
+		Keyword arguments:
+		destImage -- the destination image to write the copied metadata back to
+		"""
+		for key in self.exifKeys():
+			destImage[key] = self[key]
+		for key in self.iptcKeys():
+			destImage[key] = self[key]
+		destImage.setComment(self.getComment())
+
 def _test():
 	print 'testing library pyexiv2...'
 	# TODO: various tests
